(*
 * Copyright (C) 2006-2011 Citrix Systems Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation; version 2.1 only. with the special
 * exception on linking described in file LICENSE.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *)
module D=Debug.Make(struct let name="xapi" end)
open D

(* Mutex to prevent duplicate VGPUs being created by accident *)
let m = Mutex.create ()

(* Only allow device = "0" for now, as we support just a single vGPU per VM *)
let valid_device device =
  device = "0"

let create' ~__context  ~vM ~gPU_group ~device ~other_config ~_type ~powerstate_check =
  let vgpu = Ref.make () in
  let uuid = Uuid.to_string (Uuid.make_uuid ()) in
  if not (Pool_features.is_enabled ~__context Features.GPU) then
    raise (Api_errors.Server_error (Api_errors.feature_restricted, []));
  if powerstate_check then
    Xapi_vm_lifecycle.assert_initial_power_state_is ~__context ~self:vM ~expected:`Halted;
  if not(valid_device device) then
    raise (Api_errors.Server_error (Api_errors.invalid_device, [device]));

  (* For backwards compatibility, convert Ref.null into the passthrough type. *)
  let _type =
    if _type = Ref.null
    then Xapi_vgpu_type.find_or_create ~__context Xapi_vgpu_type.passthrough_gpu
    else begin
      if Db.is_valid_ref __context _type
      then _type
      else raise (Api_errors.Server_error
                    (Api_errors.invalid_value, ["type"; Ref.string_of _type]))
    end
  in

  Stdext.Threadext.Mutex.execute m (fun () ->
      (* Check to make sure the device is unique *)
      let all = Db.VM.get_VGPUs ~__context ~self:vM in
      let all_devices = List.map (fun self -> Db.VGPU.get_device ~__context ~self) all in
      if List.mem device all_devices then
        raise (Api_errors.Server_error (Api_errors.device_already_exists, [device]));

      Db.VGPU.create ~__context ~ref:vgpu ~uuid ~vM ~gPU_group ~device
        ~currently_attached:false ~other_config ~_type ~resident_on:Ref.null
        ~scheduled_to_be_resident_on:Ref.null
        ~compatibility_metadata:[]
        ;
    );
  debug "VGPU ref='%s' created (VM = '%s', type = '%s')" (Ref.string_of vgpu) (Ref.string_of vM) (Ref.string_of _type);
  vgpu

(* - create is defined by the autogenerated code, so we keep the same signature for it but add
   a new function create' that will accept extra parameters indicating the desired behaviour.
   - create may be called during VM.import(eg. VM cross pool migration with checkpoints), no need
   to constraint the power state in this case.
 *)
let create ~__context  ~vM ~gPU_group ~device ~other_config ~_type =
  let powerstate_check = not (Db.VM.get_is_a_snapshot ~__context ~self:vM) in
  create' ~__context  ~vM ~gPU_group ~device ~other_config ~_type ~powerstate_check

let destroy ~__context ~self =
  let vm = Db.VGPU.get_VM ~__context ~self in
  if Helpers.is_running ~__context ~self:vm then
    raise (Api_errors.Server_error (Api_errors.operation_not_allowed, ["vGPU currently attached to a running VM"]));
  Db.VGPU.destroy ~__context ~self

let atomic_set_resident_on ~__context ~self ~value = assert false

let copy ~__context ~vm vgpu =
  let all = Db.VGPU.get_record ~__context ~self:vgpu in
  let vgpu = create' ~__context
      ~device:all.API.vGPU_device
      ~gPU_group:all.API.vGPU_GPU_group
      ~vM:vm
      ~other_config:all.API.vGPU_other_config
      ~_type:all.API.vGPU_type
      ~powerstate_check:false
  in
  if all.API.vGPU_currently_attached then
    Db.VGPU.set_currently_attached ~__context ~self:vgpu ~value:true;
  vgpu

let requires_passthrough ~__context ~self =
  let _type = Db.VGPU.get_type ~__context ~self in
  Xapi_vgpu_type.requires_passthrough ~__context ~self:_type
